#include <gecode/driver.hh>
#include <gecode/int.hh>
#include <gecode/float.hh>
#include <gecode/minimodel.hh>
#include <gecode/search.hh>
#include <sys/time.h>
#include <time.h>
#include <iostream>
#include <vector>
#include <algorithm>

#include <string>

#include<gecode/gist.hh>

#include "FaultLocalization.h"
#include "TransactionData.h"

using namespace Gecode;
using namespace std;

typedef std::vector<int> VECELE;
typedef std::vector<VECELE> MATELE;

TransactionData* instance;  // Instance de la classe TransactionData
FaultLocalization* faultLoc;

int nbItems, nbTrans_p, nbTrans_m, nbTrans, compteur = 0; // Globals	variables
double tarantulaMeasure;

unsigned int nb_freq_patterns = 0; // To start to permute from (k+1)eme solution
unsigned int k = 0; 							// Top "k"

MATELE Bplus, Bmoins; 				// Bases transactionelles B+, B-

MATELE topK_patterns; 			// To store Top k solutions
VECELE sol_pattern_tmp; // Temporary vector to copy a solution on the list (List_of_solutions)
VECELE topK_freq_p; 			// To save Topk's freq+
VECELE topK_freq_m; 			// To save Topk's freq-

std::vector<double> metric; // // To store tarantula measure
string metric_name;
MATELE topK_T_plus; 	// To store Top k T+ solutions
MATELE topK_T_moins; 	// To store Top k T- solutions

bool toK_over_T; // To decide whether to calc toK over T or T-

/**
 * ScriptOption customizes the output print of Gecode
 * to Latex table format.
 *
 */

//---------------------------------------------------------
// Metrics data structures
//---------------------------------------------------------
std::vector<std::pair<int, double> > tar_metric_v;
std::vector<std::pair<int, double> > gp13_metric_v;
std::vector<std::pair<int, double> > ochiai_metric_v;
std::vector<std::pair<int, double> > ochiai2_metric_v;

bool pairCompare(const std::pair<int, double> firstElt,
		const std::pair<int, double> secondElt) {
	return firstElt.second > secondElt.second;
}

//TODO
static double _tara(double F, double P, double ef, double ep) {
	return (ef / F) / ((ef / F) + (ep / P));
}

static double _ochiai(double F, double ef, double ep) {
	return ef / sqrt((ef + ep) * (F));
}

static double _ochiai2(double F, double P, double ef, double ep) {
	return ef * (P - ep)
			/ sqrt(
					(ef + ep) * (F - ef + P - ep) * (ef + P - ep)
							* (ep + F - ef));
}
//---------------------------------------------------------
//---------------------------------------------------------
//---------------------------------------------------------
double ochiai(double ef, double ep) {
	return _ochiai((double) nbTrans_m, ef, ep);
}
double ochiai2(double ef, double ep) {
	return _ochiai2((double) nbTrans_m, (double) nbTrans_p, ef, ep);
}
double gp13(double ef, double ep) {
	return ef * (1 + (1 / ((2 * ep) + ef)));
}
double tara(double ef, double ep) {
	if ((ef + ep) == 0)
		return 0;
	return _tara((double) nbTrans_m, (double) nbTrans_p, ef, ep);
}

void get_all_metrics() {
	for (int j = 0; j < nbItems; j++) {
		int ef = 0, ep = 0;
		for (int i = 0; i < nbTrans_p; i++) {
			if (Bplus[i][j] == 1) {
				ep = ep + 1;
			}
//			else {
//				ef = ef + 1;
//			}
		}
		for (int i = 0; i < nbTrans_m; i++) {
			if (Bmoins[i][j] == 1) {
				ef = ef + 1;
			}
//			else {
//				ep = ep + 1;
//			}
		}

		tar_metric_v.push_back(std::make_pair(j, tara(ef, ep)));
		gp13_metric_v.push_back(std::make_pair(j, gp13(ef, ep)));
		ochiai_metric_v.push_back(std::make_pair(j, ochiai(ef, ep)));
		ochiai2_metric_v.push_back(std::make_pair(j, ochiai2(ef, ep)));
	}

	/*
	 cout << "BEFORE SORTING " << endl;
	 for (unsigned int i = 0; i < tar_metric_v.size(); i++) {
	 cout << "item " << tar_metric_v.at(i).first << " has tar value "
	 << tar_metric_v.at(i).second << endl;
	 }*/

}

std::string intToString(int i) {
	std::ostringstream oss;
	oss << i;
	return oss.str();
}

std::string doubleToString(double i) {
	std::ostringstream oss;
	oss << i;
	return oss.str();
}

string computeExamScore(int fault, int totalStatement,
		std::vector<std::pair<int, double> > & metric_v) { // n° of faulty statement , total number of statements of this version
	// get faulty statement score
	double faulty_stmt_score = -1;
	for (unsigned int i = 0; i < metric_v.size(); i++) {
		if (metric_v.at(i).first == fault - 1) {
			faulty_stmt_score = metric_v.at(i).second;
			break;
		}
	}
	int deltaRank = 0;
	int O_rank, P_rank;

	// Sorting
	std::sort(metric_v.begin(), metric_v.end(), pairCompare);

	// Get position of fault in ranking = O_rank
	for (unsigned int i = 0; i < metric_v.size(); i++) {
		if (metric_v[i].second == faulty_stmt_score) {
			O_rank = i + 1;
			break;
		}
	}
	// Compute P_rank
	int k = O_rank;
	while (metric_v[k].second == metric_v[O_rank - 1].second) {
		k++;
		deltaRank++;
	}
	P_rank = O_rank + deltaRank;
	// Compute ExamScores
	double O_Exam = (double) (((double) O_rank / (double) totalStatement) * 100);
	double P_Exam = (double) (((double) P_rank / (double) totalStatement) * 100);
	// print
	return intToString(P_rank) + ", " + intToString(O_rank) + ", "
			+ doubleToString(P_Exam) + ", " + doubleToString(O_Exam);
// metric_v[O_rank - 1].second
}

void computeExamScoreOf_all_metrics(int fault, int totalStatement,
		string res_Dir, string dataset_Dir) {

	string file_in_dataset_Dir = dataset_Dir;
	std::replace(file_in_dataset_Dir.begin(), file_in_dataset_Dir.end(), '/',
			'_');

	ofstream flux((res_Dir + "/Metrics_Fault_localization.txt").c_str(), //+ file_in_dataset_Dir
			ios::app);
// #{Tar_p_rank, Tar_o_rank,gp13_p_rank,gp13_o_rank,ochiai_p_rank,ochiai_o_rank,ochiai2_p_rank,ochiai2_o_rank}
	string res_row = "{";
	res_row = res_row + computeExamScore(fault, totalStatement, tar_metric_v)
			+ ",";
	res_row = res_row + computeExamScore(fault, totalStatement, gp13_metric_v)
			+ ",";
	res_row = res_row + computeExamScore(fault, totalStatement, ochiai_metric_v)
			+ ",";
	res_row = res_row
			+ computeExamScore(fault, totalStatement, ochiai2_metric_v) + "},";
	flux << res_row;
	flux.close();
}

int test_case;
enum Test_case_enum {
	Tarantula, Ochiai, Ochiai2, GP13
};

class CPBugMiner: public Space {
private:

	IntVarArray itemset; 	// Patterns

	IntVarArray Tplus;			// Transactions T+
	IntVarArray Tmoins;			// Transactions T-

	IntVar freq_plus;	// Cout à minimiser
	IntVar freq_moins;	// Cout à maximiser

	FloatVar tarantulaCoef_var;
	FloatVar ochiaiCoef_var;
	FloatVar ochiai2Coef_var;
	FloatVar GP13Coef_var;

	FloatVar f_freq_plus;
	FloatVar f_freq_moins;
	FloatVar numerateur;
	FloatVar denominateur;
	FloatVar ochiai_denominateur;

	IntVar pattern_size;

public:

// current test case : Ochiai2
	CPBugMiner(const Options& opt) {

		//---------------- INIT ------------------------//
		itemset = IntVarArray(*this, nbItems, 0, 1);
		Tplus = IntVarArray(*this, nbTrans_p, 0, 1);
		Tmoins = IntVarArray(*this, nbTrans_m, 0, 1);
		freq_plus = IntVar(*this, 0, nbTrans_p);
		freq_moins = IntVar(*this, 0, nbTrans_m);
		pattern_size = IntVar(*this, 0, nbItems);

		//---------------------------------------------------------
		// removing: obsolete intermediate structures
		//---------------------------------------------------------
		// Table de transactions sous forme IntArgs
		IntArgs Bp(nbTrans_p * nbItems);
		IntArgs Bm(nbTrans_m * nbItems);

		Matrix<IntArgs> Dplus(Bp, nbTrans_p, nbItems);
		Matrix<IntArgs> Dmoins(Bm, nbTrans_m, nbItems);

		// Calculer (1-Dti+)
		for (int i = 0; i < nbTrans_p; i++) {
			for (int j = 0; j < nbItems; j++) {
				Dplus(i, j) = 1 - Bplus[i][j];
			}
		}
		// Calculer (1-Dti-)
		for (int i = 0; i < nbTrans_m; i++) {
			for (int j = 0; j < nbItems; j++) {
				Dmoins(i, j) = 1 - Bmoins[i][j];
			}
		}

		//---------------- CONTRAINTE DE COUVERTURE ------------------------//
		//------------------------------------------------------------------//
		// Base +
		for (int i = 0; i < nbTrans_p; i++) {
			//---------------------------------------------------------
			// changing:
			//---------------------------------------------------------
			rel(*this,
					Tplus[i] == expr(*this, sum(Dplus.col(i), itemset) == 0));
		}
		// Base -
		for (int i = 0; i < nbTrans_m; i++) {
			//---------------------------------------------------------
			// changing:
			//---------------------------------------------------------
			rel(*this,
					Tmoins[i] == expr(*this, sum(Dmoins.col(i), itemset) == 0));
		}

		//---------------- CONTRAINTE DE CLOTURE ------------------------//
		// X clos <=> Il n'existe pas Y : (Y ⊃ X) ∧ (freq+(X) = freq+(Y)
		// ∧ freq-(X) = freq-(Y))
		//---------------------------------------------------------------//

		if (toK_over_T) {
			cout << ">>>> adding Cover constraint over T" << endl;
			for (int i = 0; i < nbItems; i++) {
				BoolVar bv3(*this, 0, 1);
				BoolVar bv4(*this, 0, 1);
				linear(*this, Dplus.row(i), Tplus, IRT_EQ, 0, bv3);
				linear(*this, Dmoins.row(i), Tmoins, IRT_EQ, 0, bv4);
				//---------------------------------------------------------
				// changing:
				//---------------------------------------------------------
				rel(*this, itemset[i] == expr(*this, bv3 && bv4));
			}
		} else {
			cout << ">>>> adding Cover constraint over T-" << endl;
			for (int i = 0; i < nbItems; i++) {
				BoolVar bv3(*this, 0, 1);
				linear(*this, Dmoins.row(i), Tmoins, IRT_EQ, 0, bv3);
				rel(*this, itemset[i] == bv3);
			}
		}
		//----------------- CONTRAINTE DE FREQUENCE ------------------------//
		//------------------------------------------------------------------//
		count(*this, Tplus, 1, IRT_EQ, freq_plus);
		rel(*this, freq_plus, IRT_GQ, 0);
		count(*this, Tmoins, 1, IRT_EQ, freq_moins);
		rel(*this, freq_moins, IRT_GQ, 1);

		//---------------------------------------------------------
		// changing: elementary(M)
		//---------------------------------------------------------
		rel(*this, expr(*this, freq_moins + freq_plus), IRT_NQ, 0);
		count(*this, itemset, 1, IRT_EQ, pattern_size);
		rel(*this, pattern_size, IRT_GQ, 1);

		switch (test_case) {
		case Tarantula: {
			//---------------------------------------------------------
			// adding: Trantula constraint
			//---------------------------------------------------------
			cout << "adding: Trantula constraint" << endl;

			tarantulaCoef_var = FloatVar(*this, 0.0, 1.0);
			f_freq_plus = FloatVar(*this, 0, nbTrans_p);
			f_freq_moins = FloatVar(*this, 0, nbTrans_m);
			numerateur = FloatVar(*this, 0.0, 1.0);
			denominateur = FloatVar(*this, 0.0, Float::Limits::max);

			FloatVar f_nbTransMoins(*this, nbTrans_m, nbTrans_m);
			FloatVar denominateur_fplus(*this, 0.0, 1.0);
			FloatVar f_nbTransPlus(*this, nbTrans_p, nbTrans_p);

			channel(*this, f_freq_moins, freq_moins);
			div(*this, f_freq_moins, f_nbTransMoins, numerateur);
			channel(*this, f_freq_plus, freq_plus);
			div(*this, f_freq_plus, f_nbTransPlus, denominateur_fplus);
			denominateur = expr(*this, denominateur_fplus + numerateur);
			div(*this, numerateur, denominateur, tarantulaCoef_var);

			break;
		}

		case Ochiai: {
			cout << "adding: Ochiai constraint" << endl;

			f_freq_moins = FloatVar(*this, 0, nbTrans_m);
			f_freq_plus = FloatVar(*this, 0, nbTrans_p);

			ochiaiCoef_var = FloatVar(*this, 0.0, 1.0);
			denominateur = FloatVar(*this, 0.0, Float::Limits::max);
			ochiai_denominateur = FloatVar(*this, 0.0, Float::Limits::max);
			FloatVar f_nbTransMoins(*this, nbTrans_m, nbTrans_m);

			channel(*this, f_freq_moins, freq_moins);
			channel(*this, f_freq_plus, freq_plus);
			mult(*this, expr(*this, f_freq_moins + f_freq_plus), f_nbTransMoins,
					denominateur);
			sqrt(*this, denominateur, ochiai_denominateur);
			div(*this, f_freq_moins, ochiai_denominateur, ochiaiCoef_var);
			break;
		}

		case Ochiai2: {
			cout << "adding: Ochiai2 constraint" << endl;

			f_freq_moins = FloatVar(*this, 0, nbTrans_m);
			f_freq_plus = FloatVar(*this, 0, nbTrans_p);
			ochiai2Coef_var = FloatVar(*this, 0.0, 1.0);
			ochiai_denominateur = FloatVar(*this, 0.0, Float::Limits::max);

			sqrt(*this,
					expr(*this,
							(f_freq_moins + f_freq_plus)
									* (nbTrans_m - f_freq_moins + nbTrans_p
											- f_freq_plus)
									* (f_freq_moins + nbTrans_p - f_freq_plus)
									* (f_freq_plus + nbTrans_m - f_freq_moins)),
					ochiai_denominateur);
			channel(*this, f_freq_moins, freq_moins);
			channel(*this, f_freq_plus, freq_plus);
			div(*this, expr(*this, f_freq_moins * (nbTrans_p - f_freq_plus)),
					ochiai_denominateur, ochiai2Coef_var);
			break;
		}

		case GP13: {
			cout << "adding: GP13 constraint" << endl;
			f_freq_moins = FloatVar(*this, 0, nbTrans_m);
			f_freq_plus = FloatVar(*this, 0, nbTrans_p);
			GP13Coef_var = FloatVar(*this, 0.0, Float::Limits::max);

			FloatVar rhs_GP13(*this, 0.0, 1.0);

			channel(*this, f_freq_moins, freq_moins);
			channel(*this, f_freq_plus, freq_plus);
			div(*this, f_freq_moins,
					expr(*this, (2 * f_freq_plus) + f_freq_moins), rhs_GP13);
			rel(*this, expr(*this, f_freq_moins + rhs_GP13), FRT_EQ,
					GP13Coef_var);
		}

		}
		branch(*this, itemset, INT_VAR_DEGREE_MAX(), INT_VAL_MAX()); // Du motif le plus grand au plus petit
	}

//---------------------------------------------------------
// adding: store top k patterns
//---------------------------------------------------------
	void store_topK_patterns() {
		nb_freq_patterns++;
		// Remplir les k premiers patterns dans la liste de solutions
		if (nb_freq_patterns <= k) {
//			cout << "######## Storing solution ########" << endl;
			for (int i = 0; i < nbItems; i++) {
				sol_pattern_tmp.push_back(itemset[i].val());
			}
			topK_patterns.push_back(sol_pattern_tmp);
			sol_pattern_tmp.clear();

			// Store freq+ ,freq-
			topK_freq_p.push_back(freq_plus.val());
			topK_freq_m.push_back(freq_moins.val());

			switch (test_case) {
			case Tarantula: {
				// Storing Trantula coefficient
				metric.push_back(tara(freq_moins.val(), freq_plus.val()));
				break;
			}
			case Ochiai: {
				metric.push_back(ochiai(freq_moins.val(), freq_plus.val()));
				break;
			}
			case Ochiai2: {
				metric.push_back(ochiai2(freq_moins.val(), freq_plus.val()));
				break;
			}
			case GP13: {
				metric.push_back(calcGP13(freq_moins.val(), freq_plus.val()));
			}
			}

			/// T+
			for (int i = 0; i < nbTrans_p; i++) {
				sol_pattern_tmp.push_back(Tplus[i].val());
			}
			topK_T_plus.push_back(sol_pattern_tmp);
			sol_pattern_tmp.clear();

			/// T-
			for (int i = 0; i < nbTrans_m; i++) {
				sol_pattern_tmp.push_back(Tmoins[i].val());
			}
			topK_T_moins.push_back(sol_pattern_tmp);
			sol_pattern_tmp.clear();
		} else {
//			cout << "Storing TOP1 (AFTER CALL TO Sort)" << endl;
			/// Remplacement des freq+ et freq-
			topK_freq_p[0] = freq_plus.val();
			topK_freq_m[0] = freq_moins.val();

			/// Remplacement par le nouveau motif
			for (int i = 0; i < nbItems; i++) {
				topK_patterns[0][i] = itemset[i].val();
			}
			switch (test_case) {
			case Tarantula: {
				// Storing Trantula coefficient
				metric[0] = tara(freq_moins.val(), freq_plus.val());
				break;
			}
			case Ochiai: {
				metric[0] = ochiai(freq_moins.val(), freq_plus.val());
				break;
			}
			case Ochiai2: {
				metric[0] = ochiai2(freq_moins.val(), freq_plus.val());
				break;
			}
			case GP13: {
				metric[0] = calcGP13(freq_moins.val(), freq_plus.val());
			}
			}

			/// Remplacement par le nouveau vecteur de transactions +
			for (int i = 0; i < nbTrans_p; i++) {
				topK_T_plus[0][i] = Tplus[i].val();
			}
			/// Remplacement par le nouveau vecteur de transactions -
			for (int i = 0; i < nbTrans_m; i++) {
				topK_T_moins[0][i] = Tmoins[i].val();
			}
		}
	}

//---------------------------------------------------------
// adding: print current solution + statistics
//---------------------------------------------------------

	void printStatsGecode(const Search::Statistics& stat) {
		std::cout << "Search Statistics" << std::endl;
		std::cout << "  Propagations: " << stat.propagate << std::endl;
		std::cout << "  Nodes:        " << stat.node << std::endl;
		std::cout << "  Failures:     " << stat.fail << std::endl;
		std::cout << "  Depth:        " << stat.depth << std::endl;
		int nPeakMemoryKB = static_cast<int>((stat.memory + 1023) / 1024);
		std::cout << "  Peak Memory   " << nPeakMemoryKB << "KB" << std::endl;
		std::cout << std::endl;
		printf("\n-------------------------------------- \n");
		cout << endl;
	}

//---------------------------------------------------------
// adding: computing Suspeciousness using Ochiai2 formula
//---------------------------------------------------------
	double ochiai2(double ef, double ep) {
		return ef * (nbTrans_p - ep)
				/ sqrt(
						(ef + ep) * (nbTrans_m - ef + nbTrans_p - ep)
								* (ef + nbTrans_p - ep) * (ep + nbTrans_m - ef));
	}

//---------------------------------------------------------
// adding: computing Suspeciousness using Ochiai formula
//---------------------------------------------------------
	double calcGP13(double ef, double ep) {
		return ef * (1 + (1 / ((2 * ep) + ef)));
	}

//---------------------------------------------------------
// adding: computing Suspeciousness using Trantula formula
//---------------------------------------------------------
	double __calcTarantula(double freq_m, double freq_p) const {
		double denominateur, numerateur;
		numerateur = freq_m / (double) nbTrans_m;
		denominateur = (freq_p / (double) nbTrans_p) + numerateur;
		return (numerateur / denominateur);
	}

//---------------------------------------------------------
// changing:
//---------------------------------------------------------
	void sort_topK_patterns() {
//		cout << "Sorting K patterns.........." << endl;

		/// Sorting current top k
		if (topK_freq_p.size() >= 2) {
			//---------------------------------------------------------
			// changing: simple swap
			//---------------------------------------------------------
			for (unsigned int i = 0; i < topK_freq_p.size(); i++) {
				for (unsigned int j = i + 1; j < topK_freq_p.size(); j++) {
					// Tri par tarantulaMeasure
					if (metric[j] < metric[i]) {
						// Tarantula Measure
						iter_swap(metric.begin() + i, metric.begin() + j);
						// Freq-
						iter_swap(topK_freq_m.begin() + i,
								topK_freq_m.begin() + j);
						// Freq+
						iter_swap(topK_freq_p.begin() + i,
								topK_freq_p.begin() + j);
						// Pattern
						iter_swap(topK_patterns.begin() + i,
								topK_patterns.begin() + j);
						// T+
						iter_swap(topK_T_plus.begin() + i,
								topK_T_plus.begin() + j);
						// T-
						iter_swap(topK_T_moins.begin() + i,
								topK_T_moins.begin() + j);
					}
				}
			}
		}
	}

/// Constructor for cloning
	CPBugMiner(bool share, CPBugMiner& s) :
			Space(share, s) {
		itemset.update(*this, share, s.itemset);
		Tplus.update(*this, share, s.Tplus);
		Tmoins.update(*this, share, s.Tmoins);
		freq_plus.update(*this, share, s.freq_plus);
		freq_moins.update(*this, share, s.freq_moins);
		pattern_size.update(*this, share, s.pattern_size);

		switch (test_case) {
		case Tarantula: {
			tarantulaCoef_var.update(*this, share, s.tarantulaCoef_var);
			numerateur.update(*this, share, s.numerateur);
			denominateur.update(*this, share, s.denominateur);
			break;
		}
		case Ochiai: {
			ochiaiCoef_var.update(*this, share, s.ochiaiCoef_var);
			denominateur.update(*this, share, s.denominateur);
			ochiai_denominateur.update(*this, share, s.ochiai_denominateur);
			break;
		}
		case Ochiai2: {
			ochiai2Coef_var.update(*this, share, s.ochiai2Coef_var);
			ochiai_denominateur.update(*this, share, s.ochiai_denominateur);
			break;
		}
		case GP13: {
			GP13Coef_var.update(*this, share, s.GP13Coef_var);
		}
		}

		f_freq_plus.update(*this, share, s.f_freq_plus);
		f_freq_moins.update(*this, share, s.f_freq_moins);
	}

/// Copy during cloning
	virtual Space* copy(bool share) {
		return new CPBugMiner(share, *this);
	}

	virtual void constrain(const Space& _b) {
		const CPBugMiner& b = static_cast<const CPBugMiner&>(_b);
		if (nb_freq_patterns >= k) {
			switch (test_case) {
			case Tarantula: {
				double tarantulaCoeff = tara(topK_freq_m[0], topK_freq_p[0]);
				rel(*this, b.tarantulaCoef_var, FRT_GQ, tarantulaCoeff);
				break;
			}
			case Ochiai: {
				double ochiaiCoeff = ochiai(topK_freq_m[0], topK_freq_p[0]);
				rel(*this, b.ochiaiCoef_var, FRT_GQ, ochiaiCoeff);
				break;
			}
			case Ochiai2: {
				double ochiai2Coeff = ochiai2(topK_freq_m[0], topK_freq_p[0]);
				rel(*this, b.ochiai2Coef_var, FRT_GQ, ochiai2Coeff);
				break;
			}
			case GP13: {
				double gp13Coeff = calcGP13(topK_freq_m[0], topK_freq_p[0]);
				rel(*this, b.GP13Coef_var, FRT_GQ, gp13Coeff);
			}
			}
			printf("------------ Posting new constraint -----------------\n");
			printf("-----------------------------------------------------\n");
		}
	}
}
;

//--------------------------------------------------
//--------------------------------------------------
//--------------------------------------------------
void Resolution() {

	Options opt("/***** Frequent Suspecious itemsets ******/\n");
	opt.solutions(0);
	opt.time(60 * (60 * 1000));

	CPBugMiner * m = new CPBugMiner(opt);

//	BAB<CPBugMiner> bab(m);
	DFS<CPBugMiner> dfs(m);				// Depth first search engine

//---------------------------------------------------------
// note: uncomment this if you want to visualize solving using Gist
//---------------------------------------------------------
//	Gist::bab(m);
	delete m;
	int first_K = k;
	while (CPBugMiner* sol = dfs.next()) {

		sol->store_topK_patterns();
		first_K--;
		if (first_K <= 1)
			sol->sort_topK_patterns();
		delete sol;
	}

	m->printStatsGecode(dfs.statistics());

//	GecodeStat = e.statistics();
}
//--------------------------------------------------
//--------------------------------------------------
//--------------------------------------------------
Test_case_enum to_test_case(const std::string& s) {
	if (s == "1") {
		metric_name = "Tarantula";
		return Tarantula;
	}
	if (s == "2") {
		metric_name = "Ochiai";
		return Ochiai;
	}
	if (s == "3") {
		metric_name = "Ochiai2";
		return Ochiai2;
	}
	if (s == "4") {
		metric_name = "GP13";
		return GP13;
	}
	throw std::runtime_error("invalid conversion from text to periods");
}
//--------------------------------------------------
//--------------------------------------------------
//--------------------------------------------------
int main(int argc, char* argv[]) {

//	metric_name = "Ochiai2";
	string bPlus_filename = "Bplus";
	string bMoins_filename = "Bmoins";

	string root_benchs_Dir = "";
	string dataset_Dir = "";
	string res_Dir = "";
	int fault;
	int totalStatement;
	if (argc < 3) {
		cout
				<< "Usage: ./CP4BugMining root_benchs_Dir dataset_Dir res_Dir metric_idx topK_T_idx fault_idx"
				<< endl;

		cout << "metric_idx: 1 > Tarantula\t 2 > Ochai\t 3 > Ochai2 \t 4 > GP13"
				<< endl;
		cout << "topK_T_idx: 1 > topK over T\t 0 > topK over T-" << endl;
		return 0;
	} else {
		root_benchs_Dir = argv[1];
		dataset_Dir = argv[2];
		res_Dir = argv[3];
		test_case = to_test_case(argv[4]);
		toK_over_T = atoi(argv[5]);
		fault = atoi(argv[6]);
		totalStatement = atoi(argv[7]);
	}
	string file_in_dataset_Dir = dataset_Dir;
	std::replace(file_in_dataset_Dir.begin(), file_in_dataset_Dir.end(), '/',
			'_');

	printf("---------------------------------------------------------------\n");
	cout << "Bug Mining using " << metric_name << "..." << endl;
	printf("---------------------------------------------------------------\n");

	instance = new TransactionData();
	faultLoc = new FaultLocalization();

// Parser les bases de transactions
	instance->getData_p(root_benchs_Dir + "/" + dataset_Dir + "/Base/Bplus");// bPlus_filename
	instance->getData_m(root_benchs_Dir + "/" + dataset_Dir + "/Base/Bmoins");

	Bplus = instance->getBDD_p();
	Bmoins = instance->getBDD_m();

// Charger les nombres de transactions et items
	nbTrans_p = Bplus.size() - 1;
	nbTrans_m = Bmoins.size() - 1;
	nbItems = Bplus[0].size();
	nbTrans = nbTrans_p + nbTrans_m;
	k = nbItems;

	cout << "nbItems : " << nbItems << endl;

/// Mesure de temps d'execution
	struct timeval tbegin, tend;
	double texec = 0.;
	gettimeofday(&tbegin, NULL);

	get_all_metrics();
	computeExamScoreOf_all_metrics(fault, totalStatement, res_Dir, dataset_Dir);

	//return 0;
///---- Resolution () ----
	Resolution();

	cout << "Sorting topK patterns.........." << endl;
	for (unsigned int i = 0; i < topK_freq_p.size(); i++) {
		for (unsigned int j = i + 1; j < topK_freq_p.size(); j++) {
			// Tri par Tarantula Measure
			if (metric[j] > metric[i]) {
				// Tarantula Measure
				iter_swap(metric.begin() + i, metric.begin() + j);
				// Freq-
				iter_swap(topK_freq_m.begin() + i, topK_freq_m.begin() + j);
				// Freq+
				iter_swap(topK_freq_p.begin() + i, topK_freq_p.begin() + j);
				// Pattern
				iter_swap(topK_patterns.begin() + i, topK_patterns.begin() + j);
				// T+
				iter_swap(topK_T_plus.begin() + i, topK_T_plus.begin() + j);
				// T-
				iter_swap(topK_T_moins.begin() + i, topK_T_moins.begin() + j);
			}
		}
	}

//-------------------- FAULT LOCALIZATION ----------------------------//
	faultLoc->print_topK(topK_patterns, topK_freq_p, topK_freq_m, metric,
			res_Dir, dataset_Dir, metric_name);

//	faultLoc->merge_and_rank_topK_adhoc(topK_patterns, topK_freq_p, topK_freq_m,
//			metric, res_Dir, dataset_Dir, metric_name);

	faultLoc->merge_rank_topk(topK_patterns, topK_freq_p, topK_freq_m, metric,
			res_Dir, dataset_Dir, metric_name);
//--------------------------------------------------------------------//

/// End timer
	gettimeofday(&tend, NULL);
	texec = ((double) (1000 * (tend.tv_sec - tbegin.tv_sec)
			+ ((tend.tv_usec - tbegin.tv_usec) / 1000))) / 1000.;

	ofstream fluxTM((res_Dir + "/temps_" + file_in_dataset_Dir).c_str(),
			ios::app);
	fluxTM << " Temps d'execution =  " << texec << " s." << endl;
	fluxTM.close();
	return 0;

}

