#include "FaultLocalization.h"
#include <algorithm>

typedef std::vector<int> VECELE;
typedef std::vector<VECELE> MATELE;

using namespace std;

//post::post() :
//{
//}

int FaultLocalization::print_topK(MATELE& vect_topk, VECELE& vect_fplus,
		VECELE& vect_fmoins, std::vector<double>& susp_metric, string res_Dir,
		string dataset_Dir, string metric_name) {

	string file_in_dataset_Dir = dataset_Dir;
	std::replace(file_in_dataset_Dir.begin(), file_in_dataset_Dir.end(), '/',
			'_');

	ofstream flux((res_Dir + "/Top_k_" + file_in_dataset_Dir).c_str(),
			ios::app);

	for (unsigned int i = 0; i < vect_topk.size(); i++) {
		for (vector<int>::const_iterator j = vect_topk[i].begin();
				j != vect_topk[i].end(); ++j)
			flux << *j << "";

		flux << "| Freq+ = " << vect_fplus[i] << " ,Freq- = " << vect_fmoins[i]
				<< " " << metric_name << " = " << susp_metric[i] << endl;
	}
	flux.close();
	return 0;
}

VECELE computeDeltaD(VECELE& Omega2, VECELE& SMi) {
	VECELE deltaD00;
	//TODO
	for (unsigned int i = 0; i < Omega2.size(); i++) {
		if ((Omega2[i] == 1) && (SMi[i] == 0)) {
			deltaD00.push_back(i);
		}
	}
	return deltaD00;
}

VECELE computeDeltaA(VECELE& SMi, VECELE& SM_i_moins_1) {
	VECELE deltaA00;
	//TODO
	for (unsigned int i = 0; i < SMi.size(); i++) {
		if ((SMi[i] == 1) && (SM_i_moins_1[i] == 0)) {
			deltaA00.push_back(i);
		}
	}
	return deltaA00;
}
//--------------------------------------------------
//--------------------------------------------------
//--------------------------------------------------
// Fonction qui fusionne les Si de meme frequences et classe les instructions
int FaultLocalization::merge_and_rank_topK_adhoc(MATELE& vect_topk,
		VECELE& fplus, VECELE& fmoins, std::vector<double>& metricCoef,
		string res_Dir, string dataset_Dir, string metric_name) {

	string file_in_dataset_Dir = dataset_Dir;
	std::replace(file_in_dataset_Dir.begin(), file_in_dataset_Dir.end(), '/',
			'_');

	ofstream flux1((res_Dir + "/Top-k-Fusion_" + file_in_dataset_Dir).c_str(),
			ios::app);
	ofstream flux2((res_Dir + "/Ranking_" + file_in_dataset_Dir).c_str(),
			ios::app);

	//--------------------------------------------------
	//--------------------------------------------------
	// 					Partie fusion
	//--------------------------------------------------
	//--------------------------------------------------
	unsigned int i = 0;
	unsigned int j = 0;
	MATELE SM;
	VECELE Freq_moins;
	VECELE Freq_plus;

	while (i < vect_topk.size()) {
		j = i + 1;
		VECELE tmpPattern(vect_topk[i]);
		while ((fplus[i] == fplus[j]) && (fmoins[i] == fmoins[j])) {
			for (unsigned int k = 0; k < vect_topk[i].size(); k++) {
				tmpPattern[k] = std::max(tmpPattern[k], vect_topk[j][k]);
			}
			j++;
		}
		SM.push_back(tmpPattern); 	// Les patterns
		Freq_plus.push_back(fplus[i]);		// Les F+
		Freq_moins.push_back(fmoins[i]); 	// Les F-
		tmpPattern.clear();
		i = j; // Commencer la ou les frequences ont changées

	}
	// Affichage de la nouvelle liste S' de patterns
	for (unsigned int l = 0; l < SM.size(); l++) {
		for (unsigned int jj = 0; jj < SM[l].size(); jj++) {
			flux1 << SM[l][jj] << "";
		}
		flux1 << "| Freq+ = " << Freq_plus[l] << " ,Freq- = " << Freq_moins[l]
				<< " " << metric_name << " = " << metricCoef[l] << endl;
	}

	//--------------------------------------------------
	// changing: Partie Ranking
	//--------------------------------------------------
	VECELE Omega1(SM[0].size(), 0);
	VECELE Omega2(SM[0]); // Omega2 <--- SM1
	VECELE Omega3(SM[0].size(), 0);
	VECELE _Omega1;
	VECELE _Omega3;

	for (unsigned int i = 1; i < SM.size(); i++) {
		VECELE deltaD(computeDeltaD(Omega2, SM[i]));
		if (!deltaD.empty()) {
			for (unsigned int k = 0; k < deltaD.size(); k++) {
				Omega1[deltaD[k]] = 1;
				Omega2[deltaD[k]] = 0;
				_Omega1.push_back(deltaD[k]);
			}
			_Omega1.push_back(-1); //rank delimiter
		}
		VECELE deltaA(computeDeltaA(SM[i], SM[i - 1]));
		if (!deltaA.empty()) {
			bool b = false;
			for (unsigned int k = 0; k < deltaA.size(); k++) {
				if ((Omega1[deltaA[k]] == 0) && (Omega3[deltaA[k]] == 0)) {
					b = true;
					Omega3[deltaA[k]] = 1;
					_Omega3.push_back(deltaA[k]);
				}
			}
			if (b)
				_Omega3.push_back(-1); //rank delimiter
		}
	}
	//----------------------------------------------
	VECELE Loc(_Omega1);
	Loc.push_back(-1);
	for (unsigned int i = 0; i < Omega2.size(); i++) {
		if (Omega2[i] != 0)
			Loc.push_back(i);
	}
	Loc.push_back(-1);
	Loc.insert(Loc.end(), _Omega3.begin(), _Omega3.end());

/// Afficher les differences qui sont les instructions suspectes de facon ordonée
	for (unsigned int i = 0; i < Loc.size(); i++) {
		if (Loc[i] == -1) {
			flux2 << "----" << endl;
		} else {
			flux2 << Loc[i] << endl;
		}
	}

	flux1.close();
	flux2.close();
	return 0;
}

//--------------------------------------------------
//--------------------------------------------------
//--------------------------------------------------
void FaultLocalization::merge_rank_topk(std::vector<std::vector<int> >& top_k,
		VECELE& fplus, VECELE& fmoins, std::vector<double>& metricCoef,
		string res_Dir, string dataset_Dir, string metric_name) {

	//--------------------------------------------------
	//--------------------------------------------------
	// 					Partie fusion
	//--------------------------------------------------
	//--------------------------------------------------
	unsigned int i = 0;
	unsigned int j = 0;
	MATELE SM;
	VECELE Freq_moins;
	VECELE Freq_plus;

	while (i < top_k.size()) {
		j = i + 1;
		VECELE tmpPattern(top_k[i]);
		while (metricCoef[i] == metricCoef[j]) {
			for (unsigned int k = 0; k < top_k[i].size(); k++) {
				tmpPattern[k] = std::max(tmpPattern[k], top_k[j][k]);
			}
			j++;
		}
		SM.push_back(tmpPattern); 	// Les patterns
		Freq_plus.push_back(fplus[i]);		// Les F+
		Freq_moins.push_back(fmoins[i]); 	// Les F-
		tmpPattern.clear();
		i = j; // Commencer la ou les frequences ont changées
	}

	//----------------------------------------------
	// Print merged topK
	//----------------------------------------------
	string file_in_dataset_Dir = dataset_Dir;
	std::replace(file_in_dataset_Dir.begin(), file_in_dataset_Dir.end(), '/',
			'_');
	ofstream flux1((res_Dir + "/Top-k-Fusion_" + file_in_dataset_Dir).c_str(),
			ios::app);

	ofstream flux((res_Dir + "/Ranking_" + file_in_dataset_Dir).c_str(),
			ios::app);

	// Affichage de la nouvelle liste S' de patterns
	for (unsigned int l = 0; l < SM.size(); l++) {
		for (unsigned int jj = 0; jj < SM[l].size(); jj++) {
			flux1 << SM[l][jj] << "";
		}
		flux1 << "| Freq+ = " << Freq_plus[l] << " ,Freq- = " << Freq_moins[l]
				<< " " << metric_name << " = " << metricCoef[l] << endl;
	}
	//--------------------------------------------------
	// changing: Partie Ranking
	//--------------------------------------------------
	VECELE flags(SM[0]); // 1: already ranked stmt
	VECELE ranking_topk;

	for (int j = 0; j < SM[0].size(); j++) {
		if (SM[0][j] == 1) {
			ranking_topk.push_back(j); // rank the statement
		}
	}
	ranking_topk.push_back(-1);
	for (int i = 1; i < SM.size(); i++) { // Patterns 2..k
		bool separator = false;

		for (int j = 0; j < SM[i].size(); j++) {
			if (SM[i][j] == 1 && flags[j] == 0) {
				ranking_topk.push_back(j + 1);
				flags[j] = 1; // set flags
				separator = true;  // to add separator
			}
		}
		if (separator) // insert sepator (-1) when we pass to next pattern
			ranking_topk.push_back(-1);
		else
			break;
	}
	//----------------------------------------------
	// Print topK suspecious instructions
	//----------------------------------------------

	for (unsigned int k = 0; k < ranking_topk.size(); k++) {
		if (ranking_topk[k] == -1)
			flux << "----" << endl;
		else
			flux << ranking_topk[k] << endl;
	}
	flux.close();
	//------------------

}
